@@ -122,6 +122,13 @@ static __attribute__((constructor(103))) void on_dso_load_osmo_io(void)
 	osmo_iofd_init();
 }
 
+/*! Obtain the osmo_io_backend in use by the process
+ *  \returns The osmo_io backend which was configured at startup time */
+enum osmo_io_backend osmo_io_get_backend(void)
+{
+	return g_io_backend;
+}
+
 /*! Allocate the msghdr.
  *  \param[in] iofd the osmo_io file structure
  *  \param[in] action the action this msg(hdr) is for (read, write, ..)
@@ -131,28 +138,38 @@ static __attribute__((constructor(103))) void on_dso_load_osmo_io(void)
 struct iofd_msghdr *iofd_msghdr_alloc(struct osmo_io_fd *iofd, enum iofd_msg_action action, struct msgb *msg,
 				      size_t cmsg_size)
 {
-	bool free_msg = false;
 	struct iofd_msghdr *hdr;
+	uint8_t idx, io_len;
 
-	if (!msg) {
-		msg = iofd_msgb_alloc(iofd);
-		if (!msg)
-			return NULL;
-		free_msg = true;
-	} else {
-		talloc_steal(iofd, msg);
-	}
 
 	hdr = talloc_zero_size(iofd, sizeof(struct iofd_msghdr) + cmsg_size);
-	if (!hdr) {
-		if (free_msg)
-			talloc_free(msg);
+	if (!hdr)
 		return NULL;
-	}
 
 	hdr->action = action;
 	hdr->iofd = iofd;
-	hdr->msg = msg;
+
+	/* Allocate the number of read buffers, configured by the user. Use msg as first buffer, if not NULL.
+	 * Only READ may have multiple buffers, because packets will only be written to the first buffer. */
+	io_len = (action == IOFD_ACT_READ) ? iofd->io_read_buffers : 1;
+	for (idx = 0; idx < io_len; idx++) {
+		if (msg) {
+			talloc_steal(iofd, msg);
+			hdr->msg[idx] = msg;
+			msg = NULL;
+		} else {
+			hdr->msg[idx] = iofd_msgb_alloc(iofd);
+			if (!hdr->msg[idx])
+				break;
+		}
+	}
+	/* If at least one msgb is allocated, we can continue with only one msgb, instead of completely failing. */
+	if (idx == 0) {
+		talloc_free(hdr);
+		return NULL;
+	}
+
+	hdr->io_len = idx;
 
 	return hdr;
 }
@@ -167,36 +184,19 @@ void iofd_msghdr_free(struct iofd_msghdr *msghdr)
 	talloc_free(msghdr);
 }
 
+/*! convenience wrapper to call msgb_alloc with parameters from osmo_io_fd (of given size) */
+struct msgb *iofd_msgb_alloc2(struct osmo_io_fd *iofd, size_t size)
+{
+	size_t headroom = iofd->msgb_alloc.headroom;
+
+	OSMO_ASSERT(size + headroom <= 0xffff);
+	return msgb_alloc_headroom_c(iofd, (uint16_t)(size + headroom), (uint16_t)headroom, "osmo_io_msgb");
+}
+
 /*! convenience wrapper to call msgb_alloc with parameters from osmo_io_fd */
 struct msgb *iofd_msgb_alloc(struct osmo_io_fd *iofd)
 {
-	uint16_t headroom = iofd->msgb_alloc.headroom;
-
-	OSMO_ASSERT(iofd->msgb_alloc.size <= 0xffff - headroom);
-	return msgb_alloc_headroom_c(iofd, iofd->msgb_alloc.size + headroom, headroom, "osmo_io_msgb");
-}
-
-/*! return the pending msgb in iofd or NULL if there is none*/
-struct msgb *iofd_msgb_pending(struct osmo_io_fd *iofd)
-{
-	struct msgb *msg = NULL;
-
-	msg = iofd->pending;
-	iofd->pending = NULL;
-
-	return msg;
-}
-
-/*! Return the pending msgb or allocate and return a new one */
-struct msgb *iofd_msgb_pending_or_alloc(struct osmo_io_fd *iofd)
-{
-	struct msgb *msg = NULL;
-
-	msg = iofd_msgb_pending(iofd);
-	if (!msg)
-		msg = iofd_msgb_alloc(iofd);
-
-	return msg;
+	return iofd_msgb_alloc2(iofd, iofd->msgb_alloc.size);
 }
 
 /*! Enqueue a message to be sent.
@@ -247,21 +247,52 @@ void iofd_txqueue_enqueue_front(struct osmo_io_fd *iofd, struct iofd_msghdr *msg
  */
 struct iofd_msghdr *iofd_txqueue_dequeue(struct osmo_io_fd *iofd)
 {
-	struct llist_head *lh;
+	struct iofd_msghdr *msghdr;
 
 	if (iofd->tx_queue.current_length == 0)
 		return NULL;
 
-	lh = iofd->tx_queue.msg_queue.next;
+	msghdr = llist_first_entry_or_null(&iofd->tx_queue.msg_queue, struct iofd_msghdr, list);
 
-	OSMO_ASSERT(lh);
+	OSMO_ASSERT(msghdr);
 	iofd->tx_queue.current_length--;
-	llist_del(lh);
+	llist_del(&msghdr->list);
+
+	/* Fill up empty buffers in dequeued msghdr with buffers from the next msghdr.
+	 * There can be empty buffers, when a msghdr is queued to the front with incomplete write. */
+	while (OSMO_UNLIKELY(msghdr->io_len < iofd->io_write_buffers)) {
+		struct iofd_msghdr *next;
+		int i;
+
+		if (iofd->tx_queue.current_length == 0)
+			break;
+		next = llist_first_entry_or_null(&iofd->tx_queue.msg_queue, struct iofd_msghdr, list);
+		OSMO_ASSERT(next->io_len > 0);
+		/* Get first message buffer from next msghdr and store them in the dequeued one. */
+		msghdr->iov[msghdr->io_len] = next->iov[0];
+		msghdr->msg[msghdr->io_len] = next->msg[0];
+		msghdr->hdr.msg_iovlen = ++msghdr->io_len;
+		/* Remove the message buffer from the next msghdr and free, if empty. */
+		next->io_len--;
+		for (i = 0; i < next->io_len; i++) {
+			next->iov[i] = next->iov[i + 1];
+			next->msg[i] = next->msg[i + 1];
+		}
+		if (next->io_len == 0) {
+			iofd->tx_queue.current_length--;
+			llist_del(&next->list);
+			iofd_msghdr_free(next);
+		} else {
+			memset(&next->iov[next->io_len], 0, sizeof(struct iovec));
+			next->msg[next->io_len] = NULL;
+			next->hdr.msg_iovlen = --next->io_len;
+		}
+	}
 
 	if (iofd->tx_queue.current_length == 0)
 		osmo_iofd_ops.write_disable(iofd);
 
-	return llist_entry(lh, struct iofd_msghdr, list);
+	return msghdr;
 }
 
 /*! Handle segmentation of the msg. If this function returns *_HANDLE_ONE or MORE then the data in msg will contain
@@ -336,10 +367,30 @@ void iofd_handle_segmented_read(struct osmo_io_fd *iofd, struct msgb *msg, int r
 	OSMO_ASSERT(iofd->mode == OSMO_IO_FD_MODE_READ_WRITE);
 
 	if (rc <= 0) {
+		talloc_steal(iofd->msgb_alloc.ctx, msg);
 		iofd->io_ops.read_cb(iofd, rc, msg);
 		return;
 	}
 
+	/* If we have a pending message, append the received message.
+	 * If the pending message is not large enough, create a larger message. */
+	if (OSMO_UNLIKELY(iofd->pending)) {
+		if (OSMO_UNLIKELY(msgb_tailroom(iofd->pending) < msgb_length(msg))) {
+			/* Data of msg does not fit into pending message. Allocate a new message that is larger.
+			 * This implies that msgb_length(iofd->pending) + msgb_length(msg) > iofd.msgb_alloc.size. */
+			pending = iofd_msgb_alloc2(iofd, msgb_length(iofd->pending) + msgb_length(msg));
+			OSMO_ASSERT(pending);
+			memcpy(msgb_put(pending, msgb_length(iofd->pending)), msgb_data(iofd->pending),
+			       msgb_length(iofd->pending));
+			msgb_free(iofd->pending);
+			iofd->pending = pending;
+		}
+		memcpy(msgb_put(iofd->pending, msgb_length(msg)), msgb_data(msg), msgb_length(msg));
+		msgb_free(msg);
+		msg = iofd->pending;
+		iofd->pending = NULL;
+	}
+
 	do {
 		pending = NULL;
 		res = iofd_handle_segmentation(iofd, msg, &pending);
@@ -347,6 +398,7 @@ void iofd_handle_segmented_read(struct osmo_io_fd *iofd, struct msgb *msg, int r
 			/* It it expected as per API spec that we return the
 			 * return value of read here. The amount of bytes in msg is
 			 * available to the user in msg itself. */
+			talloc_steal(iofd->msgb_alloc.ctx, msg);
 			iofd->io_ops.read_cb(iofd, rc, msg);
 			/* The user could unregister/close the iofd during read_cb() above.
 			 * Once that's done, it doesn't expect to receive any more events,
@@ -371,15 +423,16 @@ void iofd_handle_segmented_read(struct osmo_io_fd *iofd, struct msgb *msg, int r
  *  \param[in] hdr serialized msghdr containing state of completed I/O */
 void iofd_handle_recv(struct osmo_io_fd *iofd, struct msgb *msg, int rc, struct iofd_msghdr *hdr)
 {
-	talloc_steal(iofd->msgb_alloc.ctx, msg);
 	switch (iofd->mode) {
 	case OSMO_IO_FD_MODE_READ_WRITE:
 		iofd_handle_segmented_read(iofd, msg, rc);
 		break;
 	case OSMO_IO_FD_MODE_RECVFROM_SENDTO:
+		talloc_steal(iofd->msgb_alloc.ctx, msg);
 		iofd->io_ops.recvfrom_cb(iofd, rc, msg, &hdr->osa);
 		break;
 	case OSMO_IO_FD_MODE_RECVMSG_SENDMSG:
+		talloc_steal(iofd->msgb_alloc.ctx, msg);
 		iofd->io_ops.recvmsg_cb(iofd, rc, msg, &hdr->hdr);
 		break;
 	default:
@@ -395,42 +448,74 @@ void iofd_handle_recv(struct osmo_io_fd *iofd, struct msgb *msg, int rc, struct
  */
 void iofd_handle_send_completion(struct osmo_io_fd *iofd, int rc, struct iofd_msghdr *msghdr)
 {
-	struct msgb *msg = msghdr->msg;
+	int idx, i;
 
-	/* Incomplete write */
-	if (rc > 0 && rc < msgb_length(msg)) {
-		/* Re-enqueue remaining data */
-		msgb_pull(msg, rc);
-		msghdr->iov[0].iov_len = msgb_length(msg);
-		iofd_txqueue_enqueue_front(iofd, msghdr);
-		return;
-	}
-
-	/* Reenqueue the complete msgb */
+	/* Re-enqueue the complete msgb. */
 	if (rc == -EAGAIN) {
 		iofd_txqueue_enqueue_front(iofd, msghdr);
 		return;
 	}
 
-	/* All other failure and success cases are handled here */
-	switch (msghdr->action) {
-	case IOFD_ACT_WRITE:
-		if (iofd->io_ops.write_cb)
-			iofd->io_ops.write_cb(iofd, rc, msg);
-		break;
-	case IOFD_ACT_SENDTO:
-		if (iofd->io_ops.sendto_cb)
-			iofd->io_ops.sendto_cb(iofd, rc, msg, &msghdr->osa);
-		break;
-	case IOFD_ACT_SENDMSG:
-		if (iofd->io_ops.sendmsg_cb)
-			iofd->io_ops.sendmsg_cb(iofd, rc, msg);
-		break;
-	default:
-		OSMO_ASSERT(0);
-	}
+	for (idx = 0; idx < msghdr->io_len; idx++) {
+		struct msgb *msg = msghdr->msg[idx];
+		int chunk;
 
-	msgb_free(msghdr->msg);
+		/* Incomplete write */
+		if (rc > 0 && rc < msgb_length(msg)) {
+			/* Keep msg with unsent data only. */
+			msgb_pull(msg, rc);
+			msghdr->iov[idx].iov_len = msgb_length(msg);
+			/* Shift all existing buffers down. */
+			if (idx) {
+				msghdr->io_len -= idx;
+				for (i = 0; i < msghdr->io_len; i++) {
+					msghdr->iov[i] = msghdr->iov[idx + i];
+					msghdr->msg[i] = msghdr->msg[idx + i];
+				}
+				for (i = 0; i < idx; i++) {
+					memset(&msghdr->iov[msghdr->io_len + i], 0, sizeof(struct iovec));
+					msghdr->msg[msghdr->io_len + i] = NULL;
+				}
+				msghdr->hdr.msg_iovlen = msghdr->io_len;
+			}
+			/* Re-enqueue remaining buffers. */
+			iofd_txqueue_enqueue_front(iofd, msghdr);
+			return;
+		}
+
+		if (rc >= 0) {
+			chunk = msgb_length(msg);
+			if (rc < chunk)
+				chunk = rc;
+		} else {
+			chunk = rc;
+		}
+
+		/* All other failure and success cases are handled here */
+		switch (msghdr->action) {
+		case IOFD_ACT_WRITE:
+			if (iofd->io_ops.write_cb)
+				iofd->io_ops.write_cb(iofd, chunk, msg);
+			break;
+		case IOFD_ACT_SENDTO:
+			if (iofd->io_ops.sendto_cb)
+				iofd->io_ops.sendto_cb(iofd, chunk, msg, &msghdr->osa);
+			break;
+		case IOFD_ACT_SENDMSG:
+			if (iofd->io_ops.sendmsg_cb)
+				iofd->io_ops.sendmsg_cb(iofd, chunk, msg);
+			break;
+		default:
+			OSMO_ASSERT(0);
+		}
+
+		msgb_free(msghdr->msg[idx]);
+		msghdr->msg[idx] = NULL;
+
+		/* The user can unregister/close the iofd during callback above. */
+		if (!IOFD_FLAG_ISSET(iofd, IOFD_FLAG_FD_REGISTERED))
+			break;
+	}
 	iofd_msghdr_free(msghdr);
 }
 
@@ -453,6 +538,8 @@ void iofd_handle_send_completion(struct osmo_io_fd *iofd, int rc, struct iofd_ms
 int osmo_iofd_write_msgb(struct osmo_io_fd *iofd, struct msgb *msg)
 {
 	int rc;
+	struct iofd_msghdr *msghdr;
+	int idx;
 
 	if (OSMO_UNLIKELY(msgb_length(msg) == 0)) {
 		LOGPIO(iofd, LOGL_ERROR, "Length is 0, rejecting msgb.\n");
@@ -461,21 +548,35 @@ int osmo_iofd_write_msgb(struct osmo_io_fd *iofd, struct msgb *msg)
 
 	OSMO_ASSERT(iofd->mode == OSMO_IO_FD_MODE_READ_WRITE);
 
-	struct iofd_msghdr *msghdr = iofd_msghdr_alloc(iofd, IOFD_ACT_WRITE, msg, 0);
-	if (!msghdr)
-		return -ENOMEM;
+	/* Always try to add msg to last msghdr. Only if it is completely filled, allocate a new msghdr.
+	 * This way all the previous meghdrs in the queue are completely filled. */
+	msghdr = llist_last_entry_or_null(&iofd->tx_queue.msg_queue, struct iofd_msghdr, list);
+	if (msghdr && msghdr->io_len < iofd->io_write_buffers) {
+		/* Add msg to existing msghdr. */
+		msghdr->msg[msghdr->io_len++] = msg;
+	} else {
+		/* Create new msghdr with msg. */
+		msghdr = iofd_msghdr_alloc(iofd, IOFD_ACT_WRITE, msg, 0);
+		if (!msghdr)
+			return -ENOMEM;
+		msghdr->hdr.msg_iov = &msghdr->iov[0];
+		msghdr->flags = MSG_NOSIGNAL;
+	}
 
-	msghdr->flags = MSG_NOSIGNAL;
-	msghdr->iov[0].iov_base = msgb_data(msghdr->msg);
-	msghdr->iov[0].iov_len = msgb_length(msghdr->msg);
-	msghdr->hdr.msg_iov = &msghdr->iov[0];
-	msghdr->hdr.msg_iovlen = 1;
+	/* Add set IO vector to msg. */
+	idx = msghdr->io_len - 1;
+	msghdr->iov[idx].iov_base = msgb_data(msg);
+	msghdr->iov[idx].iov_len = msgb_length(msg);
+	msghdr->hdr.msg_iovlen = msghdr->io_len;
 
-	rc = iofd_txqueue_enqueue(iofd, msghdr);
-	if (rc < 0) {
-		iofd_msghdr_free(msghdr);
-		LOGPIO(iofd, LOGL_ERROR, "enqueueing message failed (%d). Rejecting msgb\n", rc);
-		return rc;
+	/* Only new msghdr will be enqueued. */
+	if (msghdr->io_len == 1) {
+		rc = iofd_txqueue_enqueue(iofd, msghdr);
+		if (rc < 0) {
+			iofd_msghdr_free(msghdr);
+			LOGPIO(iofd, LOGL_ERROR, "enqueueing message failed (%d). Rejecting msgb\n", rc);
+			return rc;
+		}
 	}
 
 	return 0;
@@ -518,8 +619,8 @@ int osmo_iofd_sendto_msgb(struct osmo_io_fd *iofd, struct msgb *msg, int sendto_
 		msghdr->hdr.msg_namelen = osmo_sockaddr_size(&msghdr->osa);
 	}
 	msghdr->flags = sendto_flags;
-	msghdr->iov[0].iov_base = msgb_data(msghdr->msg);
-	msghdr->iov[0].iov_len = msgb_length(msghdr->msg);
+	msghdr->iov[0].iov_base = msgb_data(msghdr->msg[0]);
+	msghdr->iov[0].iov_len = msgb_length(msghdr->msg[0]);
 	msghdr->hdr.msg_iov = &msghdr->iov[0];
 	msghdr->hdr.msg_iovlen = 1;
 
@@ -583,8 +684,8 @@ int osmo_iofd_sendmsg_msgb(struct osmo_io_fd *iofd, struct msgb *msg, int sendms
 	}
 
 	/* build iov from msgb */
-	msghdr->iov[0].iov_base = msgb_data(msghdr->msg);
-	msghdr->iov[0].iov_len = msgb_length(msghdr->msg);
+	msghdr->iov[0].iov_base = msgb_data(msghdr->msg[0]);
+	msghdr->iov[0].iov_len = msgb_length(msghdr->msg[0]);
 	msghdr->hdr.msg_iov = &msghdr->iov[0];
 	msghdr->hdr.msg_iovlen = 1;
 
@@ -712,6 +813,17 @@ struct osmo_io_fd *osmo_iofd_setup(const void *ctx, int fd, const char *name, en
 	iofd->tx_queue.max_length = 1024;
 	INIT_LLIST_HEAD(&iofd->tx_queue.msg_queue);
 
+	iofd->io_read_buffers = 1;
+	iofd->io_write_buffers = 1;
+
+	if (osmo_iofd_ops.setup) {
+		int rc = osmo_iofd_ops.setup(iofd);
+		if (rc < 0)  {
+			osmo_iofd_free(iofd);
+			return NULL;
+		}
+	}
+
 	return iofd;
 }
 
@@ -729,6 +841,67 @@ int osmo_iofd_set_cmsg_size(struct osmo_io_fd *iofd, size_t cmsg_size)
 	return 0;
 }
 
+/*! Set the number of buffers that can be used in a single read or write operation.
+ *
+ *  If the osmo_io_fd is in OSMO_IO_FD_MODE_READ_WRITE mode, this API function can be used to tell the
+ *  osmo_io proecess how many buffers should be read or written with a single read or write operation.
+ *  This feature is supported with io_uring backend only.
+ *
+ *  \param[in] iofd the iofd file descriptor
+ *  \param[in] op the osmo_io_op (read or write) to set the number of IO buffers for
+ *  \param[in] buffers the number of IO buffer for each specified operation
+ *  \returns zero on success, a negative value on error
+ */
+int osmo_iofd_set_io_buffers(struct osmo_io_fd *iofd, enum osmo_io_op op, uint8_t buffers)
+{
+	if (iofd->mode != OSMO_IO_FD_MODE_READ_WRITE)
+		return -EINVAL;
+
+	if (buffers < 1 || buffers > IOFD_MSGHDR_IO_BUFFERS)
+		return -EINVAL;
+
+	switch (op) {
+	case OSMO_IO_OP_READ:
+		iofd->io_read_buffers = buffers;
+		break;
+	case OSMO_IO_OP_WRITE:
+		iofd->io_write_buffers = buffers;
+		break;
+	default:
+		return -EINVAL;
+	}
+	return 0;
+}
+
+/*! Set the number of SQEs that are submitted to an io_unring before completion is received.
+ *
+ *  If the io_using backend is selected, this API function can be used to tell the osmo_io process how many SQE are
+ *  scheduled in advance.
+ *  The feature is currently supports scheduling read SQEs only.
+ *
+ *  \param[in] iofd the iofd file descriptor
+ *  \param[in] op the osmo_io_op (read) to set the number of IO buffers for
+ *  \param[in] number of scheduled SQEs
+ *  \returns zero on success, a negative value on error
+ */
+int osmo_iofd_set_sqes(struct osmo_io_fd *iofd, enum osmo_io_op op, uint8_t sqes)
+{
+	if (iofd->mode != OSMO_IO_FD_MODE_READ_WRITE)
+		return -EINVAL;
+
+	if (g_io_backend != OSMO_IO_BACKEND_IO_URING)
+		return -EINVAL;
+
+	if (op != OSMO_IO_OP_READ)
+		return -EINVAL;
+
+	if (sqes < 1 || sqes > IOFD_MSGHDR_MAX_READ_SQES)
+		return -EINVAL;
+
+	iofd->u.uring.read.num_sqes = sqes;
+	return 0;
+}
+
 /*! Register the osmo_io_fd for active I/O.
  *
  *  Calling this function will register a previously initialized osmo_io_fd for performing I/O.
@@ -824,7 +997,10 @@ void osmo_iofd_txqueue_clear(struct osmo_io_fd *iofd)
 {
 	struct iofd_msghdr *hdr;
 	while ((hdr = iofd_txqueue_dequeue(iofd))) {
-		msgb_free(hdr->msg);
+		for (int idx = 0; idx < hdr->io_len; idx++) {
+			msgb_free(hdr->msg[idx]);
+			hdr->msg[idx] = NULL;
+		}
 		iofd_msghdr_free(hdr);
 	}
 }
